unit Unit1;

{ a fraction may be assigned a unique rank
  in case of the fraction t/n there is
  - a "left" child t/(t-n)
  - a "right" child (t+n)/n

  so, a fraction may be coded as a series 0f 0's (left) and 1's (right)
  an extra "1" bit left of the code marks the start.
  The rank of the fraction = code - 1

  1/1 has rank 0, internal code 1
  1/2          1,               10     2/1  has rank 2, code 11
  1/3          3                100    3/2           4       101

  By contradiction can be proved that
  - all fractions are covered
  - fractions will occur only once
  - fractions cannot be simplified , there is no common factor

  The approach is very similar to the Euclidian algorithm
  for the calculation of the greatest common divisor of two integers

  oct. 13, 2012
  DavData
  Castricum, Netherlands
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Buttons, ExtCtrls, shellApi;

type
  TForm1 = class(TForm)
    nomedit: TEdit;
    denomedit: TEdit;
    msgtext: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    rankedit: TEdit;
    nomplus: TSpeedButton;
    nomminus: TSpeedButton;
    denomplus: TSpeedButton;
    denomminus: TSpeedButton;
    rankplus: TSpeedButton;
    rankminus: TSpeedButton;
    firstrankbtn: TSpeedButton;
    Timer1: TTimer;
    lastrankbtn: TSpeedButton;
    Image1: TImage;
    parentbtn: TSpeedButton;
    leftchildbtn: TSpeedButton;
    rightchildbtn: TSpeedButton;
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormCreate(Sender: TObject);
    procedure firstrankbtnClick(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure lastrankbtnClick(Sender: TObject);
    procedure nomeditMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure nomplusMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormPaint(Sender: TObject);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure Image1Click(Sender: TObject);
    procedure parentbtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

 TallMsg = record
            color : LongInt;
            text  : string;
           end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

const maxrank : Int64 = $7ffffffffffffffe;

var numerator : dword;
    denominator : dword;
    rank : Int64;
    timercode : byte = 0;
    fractionOK : boolean = true; //false: prevents incrementing fraction
    editcode : byte = 0;

procedure msg(nr : byte);
//display message nr in msgtext on form1
const allmsgs : array[0..13] of TAllmsg =
                 ((color:$000000 ; text:''),                         //0
                  (color:$000000 ; text:'nieuwe rank berekend'),     //1
                  (color:$000000 ; text:'nieuwe breuk berekend'),    //2
                  (color:$0000ff ; text:'illegale breuk'),           //3
                  (color:$ff00ff ; text:'maximale rank bereikt'),    //4
                  (color:$ff00ff ; text:'minimale rank bereikt'),     //5
                  (color:$0000ff ; text:'kan foute breuk niet verhogen'),//6
                  (color:$ff0000 ; text:'laagste rangorde'),          //7
                  (color:$ff0000 ; text:'hoogste rangorde'),          //8
                  (color:$000000 ; text:'welkom...'),                 //9
                  (color:$ff00ff ; text:'kleinste teller bereikt'),   //10
                  (color:$ff00ff ; text:'kleinste noemer bereikt'),   //11
                  (color:$0000ff ; text:'te hoge rangorde'),          //12
                  (color:$0000ff ; text:'gemeenschappelijke factor'));//13
begin
 with form1.msgtext do
  begin
   font.Color := allmsgs[nr].color;
   caption := allmsgs[nr].text;
  end;
end;

procedure showfraction;
begin
 form1.nomedit.Text := inttostr(numerator);
 form1.denomedit.Text := inttostr(denominator);
end;

procedure showrank;
begin
 form1.rankedit.font.Color := $0;
 form1.rankedit.text := inttostr(rank);
end;

//------ calculations ------------

function gcd(a,b : dword) : dword;
//calculate gcd of a and b
var r : dword;
begin
 repeat
  r := a mod b;
  a := b;
  b := r;
 until r = 0;
 result := a;
end;

procedure rank2fraction;
//convert rank to a fraction
//rank must be valid
var b : byte;
    r : int64;
    t,n : word;
begin
 r := rank+1;            //add startbit
 b := 64;
 while (r and $8000000000000000) = 0 do
  begin
   r := r shl 1;
   dec(b);
  end;
 t := 1;
 n := 1;
 r := r shl 1;           //strip startbit
 dec(b);
 while b > 0 do
  begin
   dec(b);
   if (r and $8000000000000000) = 0 then n := t + n else t := t + n;
   r := r shl 1;
  end;
 numerator := t;
 denominator := n;
end;

procedure fraction2rank;
//calculate the rank of a fraction
//fractionOK := true if successfull
var t,n : dword;
    b,r   : Int64;
begin
 b := 1;
 t := numerator;
 n := denominator;
 r := 0;
 fractionOK := true;
 while ((t > 1) or (n > 1)) and fractionOK do
  begin
   if t > n then
    begin
     r := r or b;
     t := t - n;
    end
   else n := n - t;
   b := b shl 1;
   if b and $8000000000000000 <> 0 then fractionOK := false;
  end;
 if fractionOK then rank := (r or b) - 1; //add start bit
end;

//------- control -------------

procedure newrank;
//rank value changed
//rank edit set before
begin
 rank2fraction;
 showfraction;
 msg(2);
end;

procedure newfraction;
//fraction value changed
//fraction edits set before
var f : dword;
begin
 f := gcd(numerator,denominator);
 if f > 1 then
  begin
   msg(13);
   form1.rankedit.Font.Color := $c0c0c0;
   exit;
  end;
 fraction2rank;
 if fractionOK then
  begin
   showrank;
   msg(1);
  end
 else
  begin
   msg(3);
   form1.rankedit.Font.Color := $c0c0c0;
  end;
end;

procedure anychange(tc : byte);
//change by  +- controls
//tc is timer (action) code
begin
 case tc of
  1 : if fractionOK then      //prevent incrementing wrong fraction
       begin
        inc(numerator);
        showfraction;
        newfraction;
       end
       else msg(6);
  2 : if numerator > 1 then
       begin
        dec(numerator);
        showfraction;
        newfraction;
       end
       else msg(10);
  3 : if fractionOK then      //prevent incrementing wrong fraction
       begin
        inc(denominator);
        showfraction;
        newfraction;
       end
       else msg(6);
  4 : if denominator > 1 then
       begin
        dec(denominator);
        showfraction;
        newfraction;
       end
       else msg(11);
  5 : if rank < maxrank then
       begin
        inc(rank);
        showrank;
        newrank;
       end
       else msg(4);
  6 : if rank > 0 then
       begin
        dec(rank);
        showrank;
        newrank;
       end
       else msg(5);
 end;
end;

procedure KBchange;
//keyboard change of Edit
//editcode = editcontrol.tag   1 : numerator, 2 : denominator, 3 : rank
begin
 with form1 do
  case editcode of
   1 : begin
        if length(nomedit.text) = 0 then nomedit.text := '1';
        numerator := strtoint(form1.nomedit.text);
        newfraction;
       end;//1
   2 : begin
        if length(denomedit.text) = 0 then denomedit.text := '1';
        denominator := strtoint(denomedit.text);
        newfraction;
       end;//2
   3 : begin
        if length(rankedit.text) = 0 then rankedit.text := '0';
        rank := strtoint64(rankedit.text);
        if rank <= maxrank then newrank
         else msg(12);
       end;//3
 end;//case
end;

procedure setfirstrank;
begin
 fractionOK := true;
 numerator := 1;
 denominator := 1;
 rank := 0;
 showrank;
 showfraction;
 msg(7);
end;

procedure setlastrank;
begin
 fractionOK := true;
 rank := maxrank;
 showrank;
 numerator := 63;
 denominator := 1;
 showfraction;
 msg(8);
end;

//--------- events --------------

procedure TForm1.FormCreate(Sender: TObject);
begin
 numerator := 1;
 denominator := 1;
 rank := 0;
 showfraction;
 showrank;
 msg(9);
end;

procedure TForm1.firstrankbtnClick(Sender: TObject);
//reset all values
begin
 setfirstrank;
end;

procedure TForm1.lastrankbtnClick(Sender: TObject);
begin
 setlastrank;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
 if timercode = 0 then timer1.Enabled := false
  else
   begin
    anychange(timercode);
    if timer1.Interval >= 200 then timer1.Interval := timer1.Interval - 50;
   end;
end;

procedure TForm1.nomeditMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 timer1.Enabled := false;
 timercode := 0;//terminate
end;

procedure TForm1.nomplusMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var t : byte;
begin
 t := TEdit(sender).Tag;
 anychange(t);
 timercode := t;
 timer1.Interval := 500;
 timer1.Enabled := true;
end;

procedure TForm1.FormKeyPress(Sender: TObject; var Key: Char);
//digit filter
begin
 editcode := 0;
 if not (key in ['0'..'9',#8]) then key := #0;
 if (activecontrol is TEdit) and (key <> #0) then
  editcode := TEdit(activecontrol).Tag;
end;

procedure TForm1.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
 if editcode <> 0 then KBchange;
 editcode := 0;
end;

procedure TForm1.FormPaint(Sender: TObject);
var x1,x2,y : word;
begin
 x1 := nomedit.Left;
 x2 := x1 + nomedit.Width;
 y := (nomedit.Top + nomedit.Height + denomedit.Top) shr 1;
 with canvas do
  begin
   pen.Color := 0;
   pen.Width := 3;
   moveto(x1,y);
   lineto(x2,y);
  end; 
end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
 if (key = VK_DELETE) and (activecontrol is TEDit) then
  editcode := TEDit(activecontrol).Tag;
end;

procedure TForm1.Image1Click(Sender: TObject);
//visit DavData website
begin
 ShellExecute(0,'open','http://www.davdata.nl', nil, nil, SW_SHOWNORMAL);
end;

procedure TForm1.parentbtnClick(Sender: TObject);
//parent, child button clicks
//tag 1: left child; 2:right child ; 3: parent
var t : byte;
begin
 t := TEdit(sender).Tag;
 if (fractionOK = false) and (t <> 3) then exit;//no child of illegal fraction

 case t of
  1 : denominator := denominator + numerator;
  2 : numerator :=  denominator + numerator;
  3 : if (numerator > 1) or (denominator > 1) then
       begin
        if numerator > denominator then numerator := numerator - denominator
         else denominator := denominator - numerator;
       end
       else exit;  
 end;//case
 showfraction;
 newfraction;
end;

end.
